pub mod schema;
pub mod models;
pub mod requests;

use diesel::prelude::*;
use models::{ Measure, NewMeasure };

use crate::measures::Measurement;
use crate::measures::MeasuresType::*;
use crate::measures::InsulinType::*;
use crate::measures::ReadingType::*;

pub fn establish_connection() -> SqliteConnection {
	let database_url = "./measures.db";
	SqliteConnection::establish(&database_url).expect(&format!("Error connecting to {}", database_url))
}

impl NewMeasure<'static> {
	pub fn is_unique(&self, connection: SqliteConnection) -> bool {
		true
	}

	pub fn from(measurement: &Measurement) -> NewMeasure<'static> {
		match &measurement.measure_type {
			Glucose(t, val) => {
				NewMeasure {
					time: measurement.time,
					data_type: "glucose",
					value: *val as f32,
					option: match t {
						Automatic => "automatic",
						Manual => "manual",
						BloodTest => "blood test",
						_ => "blood",
					},
					notes: "",
					unit: "dl/mg",
				}
			},
			Ketone(val) => {
				NewMeasure {
					time: measurement.time,
					data_type: "ketone",
					value: *val as f32,
					option: "",
					notes: "",
					unit: "dl/mg",
				}
			},
			Insulin(t, val) => {
				NewMeasure {
					time: measurement.time,
					data_type: "insulin",
					value: *val,
					option: match t {
						Long => "long",
						_ => "rapid",
					},
					notes: "",
					unit: "unit",
				}
			},
			Food(val) => {
				NewMeasure {
					time: measurement.time,
					data_type: "food",
					value: *val as f32,
					option: "",
					notes: "",
					unit: "g",
				}
			},
			Sport(val) => {
				NewMeasure {
					time: measurement.time,
					data_type: "sport",
					value: *val as f32,
					option: "",
					notes: "",
					unit: "Joule",
				}
			},
			Medications(has) => {
				NewMeasure {
					time: measurement.time,
					data_type: "medication",
					value: if *has {
							1.0
						} else {
							0.0
						},
					option: "",
					notes: "",
					unit: "",
				}
			},
		}
	}
}

impl Measure {
	pub fn is_unique(&self, connection: SqliteConnection) -> bool {
		true
	}

	pub fn to_measurement(&self) -> Measurement {
		match self.data_type.as_str() {
			"ketone" => {
				Measurement {
					time: self.time,
					measure_type: Ketone(self.value as u32),
				}
			},
			"insulin" => {
				let insulin = match self.option.as_str() {
					"long" => Long,
					_ => Rapid,
				};

				Measurement {
					time: self.time,
					measure_type: Insulin(insulin, self.value),
				}
			},
			"food" => {
				Measurement {
					time: self.time,
					measure_type: Food(self.value as u32),
				}
			},
			"sport" => {
				Measurement {
					time: self.time,
					measure_type: Sport(self.value as u32),
				}
			}
			"medications" => {
				Measurement {
					time: self.time,
					measure_type: Medications(self.value != 0.0),
				}
			},
			_ => {
				let glycemia = match self.option.as_str() {
					"automatic" => Automatic,
					"manual" => Manual,
					"blood_test" => BloodTest,
					_ => Blood,
				};

				Measurement {
					time: self.time,
					measure_type: Glucose(glycemia, self.value as u32),
				}
			},
		}
	}
}
