use super::DeviceInfo;
use super::HidGlucometer;
use super::Glucometer;

use crate::measures::*;

use chrono::prelude::*;

use std::io;
use std::collections::HashMap;

pub struct Freestyle {
	gluco: HidGlucometer,
	request_id: u8,
	reply_id: u8,
	time_map: HashMap<&'static str, usize>,
	history_map: HashMap<&'static str, usize>,
	manual_map: HashMap<&'static str, usize>,
}

enum RecordType {
	ManualMeasurements = 2,
	TimeChange = 5,
}

struct Message {
	report_hid: u8,
	id: u8,
	length: u8,
	payload: Vec<u8>,
}

struct Response {
	message: String,
	checksum: String,
	status: bool,
}

struct MultiRecordsResponse {
	message: String,
	count: u32,
	checksum: String,
}

impl Response {
	pub fn message_complete(message: &String) -> bool {
		if message.contains("CMD OK") || message.contains("CMD Fail!") {
			return true;
		}

		false
	}

	fn response_status(message: &String) -> bool {
		if message.contains("CMD OK") {
			return true;
		}

		false
	}

	fn response_message(message: &String) -> String {
		match message.find("CKSM:") {
			Some(i) => message.get(0..i).unwrap().to_string(),
			None => String::from(""),
		}
	
	}

	fn response_checksum(message: &String) -> String {
		match message.find("CKSM:") {
			Some(i) => {
				let mut checksum = String::from("0");

				if message.len() >= (i + 13) {
					checksum = message.get((i + 5)..(i + 13)).unwrap().to_string();
				}

				checksum
			},
			None => String::from("0"),
		}
	}

	pub fn new(message: String) -> Self {
		Self {
			message: Response::response_message(&message),
			checksum: Response::response_checksum(&message),
			status: Response::response_status(&message),
		}
	}

	pub fn verify_checksum(&self) -> bool {
		let message_checksum = self.message.chars().map(|c| c as u32).sum::<u32>();
		let compute_checksum = u32::from_str_radix(&self.checksum, 16);

		match compute_checksum {
			Ok(value) => {
				if message_checksum == value {
					return true;
				}

				false
			},
			Err(_err) => false,
		}
	}
}

impl MultiRecordsResponse {
	pub fn new(message: String) -> Self {
		match message.trim().rfind("\r\n") {
			Some(i) => {
				if message.len() >= (i + 2) {
					let mut checksum = String::from("0");
					let mut count = String::from("0");

					let m = message.get(0..(i + 2)).unwrap().to_string();
					let end = message.get((i + 2)..).unwrap().to_string();

					let end_vec = end.split(',').collect::<Vec<&str>>();

					if end_vec.len() >= 2 {
						count = end_vec[0].to_string();
						checksum = end_vec[1].to_string();
					}

					Self {
						message: m,
						checksum: checksum.trim().to_string(),
						count: match count.trim().parse() {
							Ok(val) => val,
							Err(_err) => 0,
						},
					}
				} else {
					Self {
						message: String::from(" "),
						checksum: String::from("0"),
						count: 0,
					}
				}
			},
			None => Self {
					message: String::from(" "),
					checksum: String::from("0"),
					count: 0,
			},
		}
	}

	pub fn get_records(&self) -> Vec<Vec<String>> {
		self.message.lines()
			.map(|s| s.trim().split(',').map(String::from).collect::<Vec<String>>())
			.collect::<Vec<Vec<String>>>()
	}

	pub fn verify_checksum(&self) -> bool {
		let message_checksum = self.message.chars().map(|c| c as u32).sum::<u32>();
		let compute_checksum = u32::from_str_radix(&self.checksum, 16);

		match compute_checksum {
			Ok(value) => {
				if message_checksum == value {
					return true;
				}

				false
			},
			Err(_e) => false,
		}
	}
}

impl Freestyle {
	pub fn new(api: &hidapi::HidApi) -> Box<dyn Glucometer> {
		let mut common_map = HashMap::new();
		let mut time_map = HashMap::new();
		let mut history_map = HashMap::new();
		let mut manual_map = HashMap::new();

		common_map.insert("id", 0);
		common_map.insert("type", 1);
		common_map.insert("month", 2);
		common_map.insert("day", 3);
		common_map.insert("year", 4);
		common_map.insert("hour", 5);
		common_map.insert("minute", 6);
		common_map.insert("second", 7);

		history_map.extend(common_map.clone());
		history_map.insert("value", 13);
		history_map.insert("sensor-lifetime-minutes", 14);
		history_map.insert("errors", 15);

		manual_map.extend(common_map.clone());
		manual_map.insert("reading-type", 9);
		manual_map.insert("value", 12);
		manual_map.insert("sport-flag", 15);
		manual_map.insert("medication-flag", 16);
		manual_map.insert("rapid-acting-flag", 17);
		manual_map.insert("long-acting-flag", 18);
		manual_map.insert("custom-comments-bitfield", 19);
		manual_map.insert("double-long-acting-insulin", 23);
		manual_map.insert("food-flag", 25);
		manual_map.insert("food-carbs-grams", 26);
		manual_map.insert("errors", 28);
		manual_map.insert("double-rapid-acting-insulin", 43);

		time_map.extend(common_map.clone());
		time_map.insert("old_month", 9);
		time_map.insert("old_day", 10);
		time_map.insert("old_year", 11);
		time_map.insert("old_hour", 12);
		time_map.insert("old_minute", 13);
		time_map.insert("old_second", 14);

		Box::new(Self {
			gluco: HidGlucometer::new(api, String::from("Freestyle Libre"), 0x1A61, 0x3650),
			request_id: 0x60,
			reply_id: 0x60,
			time_map: time_map,
			history_map: history_map,
			manual_map: manual_map,
		})
	}

	pub fn info() -> DeviceInfo {
		DeviceInfo::Hid(0x1A61, 0x3650)
	}

	fn send_command(&self, id: u8, command: &[u8]) -> Result<String, io::Error> {
		let mut com_vec = command.to_vec();

		// The payload must have 62 bytes, no more or less
		com_vec.resize(62, 0);

		let mut packet = Message {
			report_hid: 0,
			id: id,
			length: command.len() as u8,
			payload: com_vec,
		};

		let mut encoded: Vec<u8> = Vec::new();

		encoded.push(packet.report_hid);
		encoded.push(packet.id);

		// Init sequence does not have length
		if packet.length == 1 && command[0] == 0 {
			encoded.push(0);
		} else {
			encoded.push(packet.length);
		}

		encoded.append(&mut packet.payload);
		self.gluco.write(&encoded.as_slice())
	}

	fn send_command_txt(&self, command: &[u8]) -> Result<String, String> {
		let mut content = String::new();

		match self.send_command(self.request_id, command) {
			Ok(_ret) => (),
			Err(_err) => return Err(String::new()),
		}

		loop {
			let message = self.read_response();
			let response = String::from_utf8(message.payload).unwrap();

			if message.id != self.reply_id {
				return Err(String::from("Wrong message id as reply"));
			}

			println!("Content capacity: {}", content.capacity());
			println!("Content len: {}", content.len());
			println!("Response: {}", response);
			println!("Reponse len: {}", response.len());
			content += &response;

			if Response::message_complete(&response) {
				break;
			}
		}

		let response = Response::new(content);

		if !response.status || !response.verify_checksum() {
			return Err(String::from("Invalid response"));
		}

		// Remove '\r\n' sequence at the end of the message
		Ok(response.message.trim().to_string())
	}

	fn get_multi_records(&self, command: &[u8]) -> Result<Vec<Vec<String>>, String> {
		let message = match self.send_command_txt(command) {
			Ok(msg) => msg,
			Err(err) => return Err(err),
		};

		if message.is_empty() {
			return Err(String::from("Multi answser is empty"));
		}

		let multirecords = MultiRecordsResponse::new(message);

		if !multirecords.verify_checksum() {
			return Err(String::from("Multi records has a wrong checksum"));
		}

		Ok(multirecords.get_records())
	}

	fn get_value_from_list(&self, list: &Vec<String>, map: &HashMap<&str, usize>, key: &str) -> u32 {
		match map.get(key) {
			Some(i) => {
				match list[*i].parse::<u32>() {
					Ok(val) => val,
					Err(_err) => 0,
				}
			},
			None => 0,
		}
	}

	fn parse_time(&self, results: &Vec<String>) -> chrono::NaiveDateTime {
		let years: i32 = self.get_value_from_list(&results, &self.history_map, "year") as i32;
		let months: u32 = self.get_value_from_list(&results, &self.history_map, "month");
		let days: u32 = self.get_value_from_list(&results, &self.history_map, "day");
		let hours: u32 = self.get_value_from_list(&results, &self.history_map, "hour");
		let minutes: u32 = self.get_value_from_list(&results, &self.history_map, "minute");
		let seconds: u32 = self.get_value_from_list(&results, &self.history_map, "second");

		NaiveDate::from_ymd(2000 + years, months, days).and_hms(hours, minutes, seconds)
	}

	fn parse_history(&self, time: &chrono::NaiveDateTime, results: &Vec<String>) -> Option<Measurement> {
		let errors: u32 = self.get_value_from_list(&results, &self.history_map, "errors");

		if errors != 0 {
			return None;
		}

		let value: u32 = self.get_value_from_list(&results, &self.history_map, "value");

		Some(Measurement {
			measure_type: MeasuresType::Glucose(ReadingType::Automatic, value),
			time: *time,
		})
	}

	fn parse_array(&self, time: &chrono::NaiveDateTime, results: &Vec<String>) -> Vec<Measurement> {
		let mut ret = Vec::new();
		let errors: u32 = self.get_value_from_list(&results, &self.manual_map, "errors");

		if errors != 0 {
			return ret;
		}

		let reading_type_auto = 0;
		let reading_type_ketone_blood = 1;
		let reading_type_glucose_blood = 2;

		let value: u32 = self.get_value_from_list(&results, &self.manual_map, "value");
		let has_rapid_acting: u32 = self.get_value_from_list(&results, &self.manual_map, "rapid-acting-flag");
		let reading_type: u32 = self.get_value_from_list(&results, &self.manual_map, "reading-type");
		let has_sport: u32 = self.get_value_from_list(&results, &self.manual_map, "sport-flag");
		let has_medic: u32 = self.get_value_from_list(&results, &self.manual_map, "medication-flag");
		let has_food: u32 = self.get_value_from_list(&results, &self.manual_map, "food-flag");
		let has_long_acting: u32 = self.get_value_from_list(&results, &self.manual_map, "long-acting-flag");

		if reading_type == reading_type_auto {
			ret.push(Measurement {
				measure_type: MeasuresType::Glucose(ReadingType::Manual, value),
				time: *time,
			})
		} else if reading_type == reading_type_glucose_blood {
			ret.push(Measurement {
				measure_type: MeasuresType::Glucose(ReadingType::Blood, value),
				time: *time,
			})
		} else if reading_type == reading_type_ketone_blood {
			ret.push(Measurement {
				measure_type: MeasuresType::Ketone(value),
				time: *time,
			})
		}

		if has_sport != 0 {
			ret.push(Measurement {
				measure_type: MeasuresType::Sport(0),
				time: *time,
			})
		}

		if has_medic != 0 {
			ret.push(Measurement {
				measure_type: MeasuresType::Medications(true),
				time: *time,
			})
		}

		if has_food != 0 {
			let food_value = self.get_value_from_list(&results, &self.manual_map, "food-carbs-grams");
			ret.push(Measurement {
				measure_type: MeasuresType::Food(food_value),
				time: *time,
			})
		}

		if has_long_acting != 0 {
			let long_value: f32 = self.get_value_from_list(&results, &self.manual_map, "double-long-acting-insulin") as f32;
			let insulin_value = long_value / 2.0;

			ret.push(Measurement {
				measure_type: MeasuresType::Insulin(InsulinType::Long, insulin_value),
				time: *time,
			})
		}

		if has_rapid_acting != 0 {
			let rapid_value: f32 = self.get_value_from_list(&results, &self.manual_map, "double-rapid-acting-insulin") as f32;
			let insulin_value = rapid_value / 2.0;

			ret.push(Measurement {
				measure_type: MeasuresType::Insulin(InsulinType::Rapid, insulin_value),
				time: *time,
			})
		}

		ret
	}

	fn parse_results(&self, results: &Vec<String>) -> Vec<Measurement> {
		let array_type = 2;
		let time_type = 5;

		let time = self.parse_time(&results);
		let measure_type: u32 = self.get_value_from_list(&results, &self.history_map, "type");
		let mut ret = Vec::new();

		if measure_type == array_type {
			ret.append(&mut self.parse_array(&time, &results));
		} else if measure_type == time_type {
			();
		} else {
			match self.parse_history(&time, &results) {
				Some(m) => ret.push(m),
				None => (),
			}
		}

		ret
	}

	fn read_response(&self) -> Message {
		let mut buf = vec![0u8; 64];

		match self.gluco.read(&mut buf[..]) {
			Ok(_i) => (),
			Err(err) => println!("Reading failed: {}", err),
		}

		let id = buf[0];
		let length = buf[1];

		// Some devices sent sync packets which can be skipped
		if id == 0x22 && length == 1 {
			return self.read_response();
		}
	
		let payload = buf[2..(2 + length as usize)].to_vec();

		Message {
			report_hid: 0,
			id: id,
			length: length,
			payload: payload,
		}
	}
}

impl Glucometer for Freestyle {
	fn open(&mut self) -> Result<String, io::Error> {
		let init_seq = vec![0x04u8, 0x05u8, 0x15u8, 0x01u8];
		let ret = self.gluco.open();

		match &ret {
			Ok(_i) => (),
			Err(_err) => return ret,
		}

		for id in init_seq.iter() {
			let command = vec![0u8; 1];

			match self.send_command(*id, &command[..]) {
				Ok(_i) => {
					self.read_response();
				},
				Err(err) => {
					println!("Writing failed: {}", err);
					break;
				},
			}
		}

		ret
	}

	fn close(&self) {
		self.gluco.close()
	}

	fn read(&self) -> Result<Vec<Measurement>, String> {
		let mut ret = Vec::new();

		let history_records = match self.get_multi_records("$history?".as_bytes()) {
			Ok(history) => history,
			Err(err) => return Err(err),
		};

		let manual_records = match self.get_multi_records("$arresult?".as_bytes()) {
			Ok(manual) => manual,
			Err(err) => return Err(err),
		};

		for line in history_records {
			ret.append(&mut self.parse_results(&line));
		}

		for line in manual_records {
			ret.append(&mut self.parse_results(&line));
		}

		Ok(ret)
	}

	fn info(&self) -> DeviceInfo {
		self.gluco.info()
	}

	fn get_patient_name(&self) -> Option<String> {
		match self.send_command_txt("$ptname?".as_bytes()) {
			Ok(name) => Some(name),
			Err(_err) => None,
		}
	}

	fn get_version(&self) -> Option<String> {
		match self.send_command_txt("$swver?".as_bytes()) {
			Ok(version) => Some(version),
			Err(_err) => None,
		}
	}

	fn get_serial_number(&self) -> Option<String> {
		match self.send_command_txt("$sn?".as_bytes()) {
			Ok(sn) => Some(sn),
			Err(_err) => None,
		}
	}

	fn get_date(&self) -> Option<chrono::NaiveDateTime> {
		let date = match self.send_command_txt("$date?".as_bytes()) {
			Ok(date) => date,
			Err(_err) => return None,
		};

		let time = match self.send_command_txt("$time?".as_bytes()) {
			Ok(time) => time,
			Err(_err) => return None,
		};

		let date_fields = date.split(',').collect::<Vec<&str>>();
		let time_fields = time.split(',').collect::<Vec<&str>>();

		let year: i32 = date_fields[2].parse().unwrap();
		let month: u32 = date_fields[0].parse().unwrap();
		let day: u32 = date_fields[1].parse().unwrap();

		let hour: u32 = time_fields[0].parse().unwrap();
		let minute: u32 = time_fields[1].parse().unwrap();

		Some(NaiveDate::from_ymd(2000 + year, month, day).and_hms(hour, minute, 0))
	}

	fn set_patient_name(&self, name: String) {
		let command = String::from("$ptname,") + &name;
		self.send_command_txt(command.as_bytes());
	}

	fn set_patient_id(&self, id: String) {
		let command = String::from("$ptid,") + &id;
		self.send_command_txt(command.as_bytes());
	}

	fn get_patient_id(&self) -> Option<String> {
		match self.send_command_txt("$ptid?".as_bytes()) {
			Ok(ptid) => Some(ptid),
			Err(_err) => None,
		}
	}

	fn get_nb_records(&self) -> Option<u32> {
		let reply = match self.send_command_txt("$dbrnum?".as_bytes()) {
			Ok(msg) => msg,
			Err(_err) => return None,
		};
		let pattern = "DB Record Number = ";

		match &reply.find(pattern) {
			Some(i) => Some(reply.get((i + pattern.len())..).unwrap().parse::<u32>().unwrap()),
			None => None,
		}
	}

	fn set_date(&self, date: chrono::NaiveDateTime) {
		let date_command = String::from("$date,") + &date.month().to_string()
			+ &String::from(",") + &date.day().to_string()
			+ &String::from(",") + &(date.year() - 2000).to_string();

		let time_command = String::from("$time,") + &date.hour().to_string()
			+ &String::from(",") + &date.minute().to_string();

		self.send_command_txt(date_command.as_bytes());
		self.send_command_txt(time_command.as_bytes());
	}

	fn get_glucose_unit(&self) -> ConcentrationUnit {
		ConcentrationUnit::MgPerDeciliter
	}

	fn get_ketone_unit(&self) -> ConcentrationUnit {
		ConcentrationUnit::MgPerDeciliter
	}
}
