pub mod driver;
mod freestyle;

use super::measures::*;

use std::io;

use hidapi::HidApi;

pub enum DeviceInfo {
	Hid(u16, u16),
}

pub trait Glucometer {
	fn open(&mut self) -> Result<String, io::Error>;
	fn close(&self);
	fn info(&self) -> DeviceInfo;
	fn read(&self) -> Result<Vec<Measurement>, String>;
	fn get_patient_name(&self) -> Option<String>;
	fn set_patient_name(&self, name: String);
	fn get_patient_id(&self) -> Option<String>;
	fn set_patient_id(&self, id: String);
	fn get_version(&self) -> Option<String>;
	fn get_serial_number(&self) -> Option<String>;
	fn get_date(&self) -> Option<chrono::NaiveDateTime>;
	fn set_date(&self, date: chrono::NaiveDateTime);
	fn get_glucose_unit(&self) -> ConcentrationUnit;
	fn get_ketone_unit(&self) -> ConcentrationUnit;
	fn get_nb_records(&self) -> Option<u32>;
}

pub struct HidGlucometer {
	name: String,
	vendor_id: u16,
	product_id: u16,
	hidapi: * const hidapi::HidApi,
	device: Option<hidapi::HidDevice>,
}

impl HidGlucometer {
	pub fn new(api: &HidApi, name: String, vendor_id: u16, product_id: u16) -> Self {
		println!("Found Hid glucometer {} with id 0x{:X?}:0x{:X?}", name, vendor_id, product_id);

		Self {
			name: name,
			hidapi: api,
			vendor_id: vendor_id,
			product_id: product_id,
			device: None,
		}
	}

	pub fn info(&self) -> DeviceInfo {
		match &self.device {
			Some(dev) =>  {
				println!("Device info: {:?}", dev.get_manufacturer_string().expect("Failure"));
				println!("Device info: {:?}", dev.get_product_string().expect("Failure"));
				println!("Device info: {:?}", dev.get_serial_number_string().expect("Failure"));
			},
			_ => (),
		}

		DeviceInfo::Hid(self.vendor_id, self.product_id)
	}

	pub fn open(&mut self) -> Result<String, io::Error> {
		unsafe {
			match (*self.hidapi).open(self.vendor_id, self.product_id) {
				Ok(dev) => {
					dev.set_blocking_mode(true);
					self.device = Some(dev);
					Ok(String::from("HID device openned"))
				},
				Err(_err) => Err(io::Error::new(io::ErrorKind::NotFound, "Device not opened")),
			}
		}
	}

	pub fn close(&self) { }

	pub fn read(&self, data: &mut [u8]) -> Result<String, io::Error> {
		match &self.device {
			Some(dev) => {
				match dev.read(&mut data[..]) {
					Ok(nb) => Ok(String::from(format!("HID device reading {} bytes", nb))),
					Err(_e) => {
						Err(io::Error::new(io::ErrorKind::InvalidInput, "HID reading failed"))
					},
				}
			},

			None => Err(io::Error::new(io::ErrorKind::NotFound, "HID device not found")),
		}
	}

	pub fn write(&self, data: &[u8]) -> Result<String, io::Error> {
		match &self.device {
			Some(dev) => {
				match dev.write(&data[..]) {
					Ok(nb) => Ok(String::from(format!("HID device writing {} bytes", nb))),
					Err(e) => {
						println!("Writing failed: {}", e);
						Err(io::Error::new(io::ErrorKind::InvalidInput, "HID writing failed"))
					},
				}
			},

			None => Err(io::Error::new(io::ErrorKind::NotFound, "HID device not found")),
		}
	}
}
