use crate::measures::*;
use crate::measures::Measurement;

use gtk::prelude::*;
use glib::clone;

use super::graphs::{HorizontalAxis, VerticalAxis, Axis, AxisType, AxisPosition, ChartWidget, Range, Point, Series, SplineSeries, DotSeries};

use gtk::{Builder, Calendar, DrawingArea, Button};

use chrono::{Local, Datelike, NaiveDate, NaiveDateTime};
use crate::database::requests::*;

fn data_to_series(data: &Vec<Measurement>, continuous: &mut dyn Series, manual: &mut dyn Series) {
	for m in data {
		match &m.measure_type {
			MeasuresType::Glucose(t, val) => {
				match t {
					ReadingType::Automatic => continuous.add_point(Point::new(m.time.timestamp() as f64, *val as f64)),
					ReadingType::Manual => manual.add_point(Point::new(m.time.timestamp() as f64, *val as f64)),
					_ => (),
				};
			},
			_ => (),
		}
	}
}

fn calendar_widget_to_date(calendar: &Calendar) -> NaiveDate {
	let date_widget = calendar.get_date();
	chrono::NaiveDate::from_ymd(date_widget.0 as i32, date_widget.1 + 1, date_widget.2)
}

fn date_to_calendar_widget(calendar: &Calendar, date: &NaiveDate) {
	calendar.select_month(date.month() - 1, date.year() as u32);
	calendar.select_day(date.day());

	calendar.clear_marks();
	calendar.mark_day(date.day());
}

pub fn build(builder: &Builder) {
	let date = NaiveDate::from_ymd(2020, 1, 27);
	let today = NaiveDateTime::from_timestamp(Local::now().timestamp(), 0).date();

	let calendar: Calendar = builder.get_object("history-calendar").expect("Couldn't get calendar");
	let next: Button = builder.get_object("history-day-next").expect("Couldn't get history-day-next");
	let previous: Button = builder.get_object("history-day-prev").expect("Couldn't get history-day-prev");
	let draw: DrawingArea = builder.get_object("history-chart").expect("Couldn't get history");
	draw.set_size_request(1000, 550);

	let chart = new_chart(draw.clone(), date.clone());

	date_to_calendar_widget(&calendar, &date);

	calendar.connect_day_selected(clone!(@weak draw => move |calendar| {
		let chart = new_chart(draw.clone(), calendar_widget_to_date(&calendar));
		chart.draw();
	}));

	next.connect_clicked(clone!(@weak draw, @weak calendar => move |next| {
		let date = calendar_widget_to_date(&calendar).succ();

		date_to_calendar_widget(&calendar, &date);
		let chart = new_chart(draw.clone(), date.clone());
		chart.draw();
	}));

	previous.connect_clicked(clone!(@weak draw, @weak calendar => move |previous| {
		let date = calendar_widget_to_date(&calendar).pred();

		date_to_calendar_widget(&calendar, &date);
		let chart = new_chart(draw.clone(), date.clone());
		chart.draw();
	}));

	chart.draw();
}

fn new_chart(draw: DrawingArea, date: NaiveDate) -> ChartWidget {
	let measures = get_measurement_from_date(date);

	let mut x_axis = HorizontalAxis::new("Time".to_string(), AxisType::Date, AxisPosition::AxisBottom, Range::new(date.and_hms(0, 0, 0).timestamp() as f64, date.and_hms(23, 59, 59).timestamp() as f64).unwrap());
	let mut y_axis = VerticalAxis::new("Glycemia".to_string(), AxisType::Real, AxisPosition::AxisLeft, Range::new(0., 500.).unwrap());

	y_axis.set_unit("mg/dl".to_string());
	y_axis.add_marker(90.);
	y_axis.add_marker(180.);

	x_axis.set_unit("h".to_string());
	x_axis.add_marker(date.and_hms(6, 0, 0).timestamp() as f64);
	x_axis.add_marker(date.and_hms(12, 0, 0).timestamp() as f64);
	x_axis.add_marker(date.and_hms(18, 0, 0).timestamp() as f64);
	let mut chart = ChartWidget::new(draw, Box::new(x_axis), Box::new(y_axis));

	let mut series_continuous = SplineSeries::new(0, String::from("Continuous"));
	let mut series_manual = DotSeries::new(1, String::from("Manual"), String::from("assets/manual-icon.png"));

	data_to_series(&measures, &mut series_continuous, &mut series_manual);

	chart.add_series(Box::new(series_continuous));
	chart.add_series(Box::new(series_manual));

	chart
}
