use cairo::{Context, FontSlant, FontWeight, Rectangle};

use chrono::NaiveDateTime;

#[derive(Debug, Copy, Clone)]
pub enum AxisType {
	Date,
	Real,
}

pub enum AxisOrientation {
	VerticalAxis,
	HorizontalAxis,
}

pub enum AxisPosition {
	AxisTop,
	AxisBottom,
	AxisLeft,
	AxisRight,
}

pub trait Axis {
	fn add_marker(&mut self, marker: f64);
	fn set_unit(&mut self, unit: String);
	fn set_step(&mut self, step: f64);
	fn set_orientation(&mut self, orientation: AxisOrientation);
	fn set_range(&mut self, range: Range);
	fn set_description(&mut self, description: String);
	fn get_legend(&self) -> String;
	fn get_scale_value(&self, range: f64) -> f64;
	fn get_range(&self) -> Range;
	fn draw(&self, context: &Context, surface: &Rectangle);
}

#[derive(Debug, Copy, Clone)]
pub struct Range {
	min: f64,
	max: f64,
}

pub struct AbstractAxis {
	pub range: Range,
	pub axis_type: AxisType,
	pub step: f64,
	pub markers: Vec<f64>,
	pub orientation: AxisOrientation,
	pub position: AxisPosition,
	pub unit: String,
	pub description: String,
}

impl Range {
	pub fn new(min: f64, max: f64) -> Option<Self> {
		if min >= max {
			return None;
		}

		Some(Range {
			min: min,
			max: max,
		})
	}

	pub fn min(&self) -> f64 {
		self.min
	}

	pub fn max(&self) -> f64 {
		self.max
	}
}

impl AbstractAxis {
	pub fn new(description: String, axis_type: AxisType, orientation: AxisOrientation, position: AxisPosition, range: Range) -> Self {
		Self {
			range: range,
			axis_type: axis_type,
			step: 1.0,
			orientation: orientation,
			position: position,
			markers: Vec::new(),
			unit: String::new(),
			description: description,
		}
	}

	pub fn add_marker(&mut self, marker: f64) {
		self.markers.push(marker);
	}

	pub fn set_unit(&mut self, unit: String) {
		self.unit = unit;
	}

	pub fn set_step(&mut self, step: f64) {
		self.step = step;
	}

	pub fn set_orientation(&mut self, orientation: AxisOrientation) {
		self.orientation = orientation;
	}

	pub fn set_range(&mut self, range: Range) {
		self.range = range;
	}

	pub fn get_range(&self) -> Range {
		self.range
	}

	pub fn get_type(&self) -> AxisType {
		self.axis_type
	}

	pub fn set_description(&mut self, description: String) {
		self.description = description;
	}

	pub fn get_legend(&self) -> String {
		match &self.axis_type {
			AxisType::Date => {
				let date = NaiveDateTime::from_timestamp(self.range.min() as i64, 0);
				format!("{} ({}) from {}", self.description, self.unit, date.format("%Y-%m-%d").to_string())
			},
			_ => {
				format!("{} ({})", self.description, self.unit)
			},
		}
	}

	pub fn get_scale_value(&self, range: f64) -> f64 {
		range / (self.range.max() - self.range.min())
	}

	pub fn draw_text(&self, context: &Context, text: &String) {
		context.select_font_face("Sans", FontSlant::Normal, FontWeight::Normal);
		context.set_font_size(20.0);

		context.show_text(text);
	}

	pub fn get_relative_value(&self, value: f64) -> f64 {
		value - self.range.min()
	}

	pub fn marker_text(&self, marker: f64) -> String {
		let rel_val = marker - self.range.min();

		match &self.axis_type {
			AxisType::Date => {
				let date = NaiveDateTime::from_timestamp(rel_val as i64, 0);
				date.format("%H:%M").to_string()
			},
			_ => {
				rel_val.to_string()
			},
		}
	}
}
