extern crate gtk;
extern crate gio;
extern crate cairo;

pub mod axis;
pub mod horizontal_axis;
pub mod vertical_axis;
pub mod series;
pub mod spline_series;
pub mod dot_series;

use gtk::prelude::*;

pub use axis::{Axis, AxisType, AxisOrientation, AxisPosition, Range};
pub use series::{Point, Series};
pub use dot_series::DotSeries;
pub use spline_series::SplineSeries;
pub use vertical_axis::VerticalAxis;
pub use horizontal_axis::HorizontalAxis;

use cairo::{Context, Format, ImageSurface, Rectangle};
use gtk::{DrawingArea};

pub struct ChartWidget {
	x: Box<dyn Axis>,
	y: Box<dyn Axis>,
	series: Vec<Box<dyn Series>>,
	chart_margin: f64,
	chart_surface: Rectangle,
	draw_area: DrawingArea,
}

impl ChartWidget {
	pub fn new(area: DrawingArea, x: Box<dyn Axis>, y: Box<dyn Axis>) -> Self {
		let size = area.get_size_request();
		let width = size.0 as f64;
		let height = size.1 as f64;
		let margin = 75.0;

		ChartWidget {
			x: x,
			y: y,
			series: Vec::new(),
			draw_area: area,
			chart_margin: margin,
			chart_surface: Rectangle {
				x: margin,
				y: height - margin,
				width: width - margin,
				height: height - margin,
			}
		}
	}

	pub fn set_axis(&mut self, orientation: AxisOrientation, axis: Box<dyn Axis>) {
		match orientation {
			AxisOrientation::VerticalAxis => self.y = axis,
			AxisOrientation::HorizontalAxis => self.x = axis,
		};
	}

	pub fn add_series(&mut self, series: Box<dyn Series>) {
		self.series.push(series);
	}

	fn get_widget_size(&self) -> Point {
		let size = self.draw_area.get_size_request();

		Point::new(size.0 as f64, size.1 as f64)
	}

	fn get_point_origin(&self) -> Point {
		let origin = self.get_widget_size();

		Point::new(origin.x() * 0.05, origin.x() * 0.05)
	}

	pub fn draw(&self) {
		let size = self.get_widget_size();
		// let mut image = ImageSurface::create(Format::Rgb24, size.x() as i32, size.y() as i32).expect("Can't create surface");
		let image = ImageSurface::create(Format::ARgb32, size.x() as i32, size.y() as i32).expect("Can't create surface");
		let cr = Context::new(&image);
		cr.set_line_width(3.0);
		cr.stroke();

		cr.set_source_rgb(0., 255., 0.);
		self.x.draw(&cr, &self.chart_surface);
		cr.set_source_rgb(0., 0., 255.);
		self.y.draw(&cr, &self.chart_surface);

		for serie in &self.series {
			let legend_x = self.chart_surface.x + self.chart_surface.width / 4. * serie.get_id() as f64;
			let legend_y = 20.;
			let legend_point = Point::new(legend_x, legend_y);

			if serie.get_id() == 1 {
				cr.set_source_rgb(255., 0., 0.);
			} else {
				cr.set_source_rgb(255., 255., 255.);
			}

			serie.draw(&cr, &self.chart_surface, &*self.x, &*self.y);
			serie.draw_legend(&cr, &self.chart_surface, legend_point);
		}

		self.draw_area.queue_draw();

		self.draw_area.connect_draw(move |_, cr| {
			cr.set_source_surface(&image, 0.0, 0.0);
			cr.paint();
			image.flush();
			image.finish();
			Inhibit(false)
		});
	}
}
